/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/


#include <libc.h>
#include <string.h>
#include <stdio.h>

#include "oo7.h"
#include "globals.h"
#include "GenParams.h"
#include "BenchParams.h"
#include "VarParams.h"
#include "PartIdList.h"
#include "Iterator.h"

//////////////////////////////////////////////////////////////////
//
// Reorganization Test #1 - Visit all composite parts, deleting
// and then newly inserting 1/2 of each composite part's atomic
// parts graph.
//
//////////////////////////////////////////////////////////////////

extern char*	types[NumTypes];
extern o_4b	debugMode;

static Iterator	partI;

//////////////////////////////////////////////////////////////////
//
// Routine to do Reorg1
//
//////////////////////////////////////////////////////////////////

o_4b reorg1()
{
    // iterate through all the composite parts, deleting/inserting
    // half of an atomic parts graph worth of parts and counting up
    // the number of atomic parts processed in this way

    o_4b		count = 0;
    CompositePart*	cp;
    PPredicate		pred = NULL_PREDICATE;

    // Iterate over all composite parts.
    partI.select(Composite, pred);
    for (cp = (Link<CompositePart>)partI.first(); cp != NULL;
	 cp = (Link<CompositePart>)partI.next())
    {
        if (debugMode) {
            printf("    In Reorg1, CompPartId = %d:\n", cp->id);
        }
        count += cp->reorg1();
	cp->unpinobj();
    }
    return count;
}

//////////////////////////////////////////////////////////////////
//
// Composite Part Methods for Reorg1
//
//////////////////////////////////////////////////////////////////

o_4b CompositePart::reorg1()
{
    AtomicPart* atomicPart;
    o_4b	partId;
    //JGM 93/10/15 Changed o_word to o_4b to avoid cfront warnings
    o_4b	i;

    dirty();

    if (debugMode) {
        printf("CompositePart::Reorg1(id = %d)\n", id);

    }

    // start by getting pointers to all of the atomic parts, in increasing
    // part id order, in preparation for removing/recreating half of them

    AtomicPart**		atom = new AtomicPart*[NumAtomicPerComp];
    o_4b			lowId = rootPart->id;
    o_4b			highId = lowId + NumAtomicPerComp - 1;
    PPredicate	pred =		PAttribute("AtomicPart::id") <= highId &&
				PAttribute("AtomicPart::id") >= lowId;
    LinkVstr<AtomicPart>	atoms = 
		PClassObject<AtomicPart>::Object().select(NULL, FALSE, pred);

    ::dom->greadobjs(atoms, NULL);
    for (i = 0, partId = lowId; i < NumAtomicPerComp; i++, partId++) 
    {
	atom[i] = atoms[i];
	if (atom[i] == NULL || atom[i]->id != partId) {
            fprintf(stderr, "ERROR: Unable to find atomic part %d\n", partId);
            exit(1);
        }
    }

    // now do the deletions of half of the atomic parts, keeping the first
    // half and targeting the second half for simplicity (to avoid the root
    // part and to enable a simple computation to discriminate between parts
    // to be deleted and those to be retained);
    // before doing the deletions, go through and remember all the connections
    // that will be lost when the targeted parts are deleted;
    // after all that, recreate the parts and reconnect them to everything!

    // first compute the range of impact

    o_4b	amtToChange = NumAtomicPerComp / 2;
    o_4b	firstToChange = NumAtomicPerComp - amtToChange;
    o_4b	midId = highId - amtToChange + 1;

    // next record the relevant connections.

    // allocate space to hold the connected parts
    PartIdList*	toParts = new PartIdList[NumAtomicPerComp];

    for (i = 0; i < NumAtomicPerComp; i++) 
    {
	// for each of the atomic parts in the atom[] array
	// scan through the parts it is connected to
        LinkVstr<Connection>&	connections = atom[i]->to;

	for (o_word j = 0; j < connections.size(); j++) {
	    Connection*	connection = connections[j];

            atomicPart = connection->to;  // get the associated atomic part
	    // if either the part or the connected part have id's
	    // greater than midId, save the id of part in the toParts 
	    // array
	    if ((atom[i]->id >= midId) || (atomicPart->id >= midId)) {
		toParts[i].insert(atomicPart->id);
	    }
	    connection->unpinobj();
        }
    }

    // now do the actual part deletions
    o_4b	count = 0;

    for (i = firstToChange; i < NumAtomicPerComp; i++) 
    {
	if (debugMode) {
            printf("\tReorg1: Deleting atomic part %d\n", atom[i]->id);
	}
	delete atom[i];  
	count++;
    }
    if (debugMode) {
	printf("\tReorg1: Deleted all atomic parts\n");
    }
    
    // time to (approximately) recreate the deleted parts
    for (i = firstToChange, partId = midId; i < NumAtomicPerComp; i++, partId++)
    {
	atom[i] = new Persistent AtomicPart(partId, this);
	if (debugMode) {
            printf("\tReorg1: Recreated atomic part %d\n", atom[i]->id);
	}
    }

    // finally, connect all of the parts back up again
    for (i = 0; i < NumAtomicPerComp; i++) 
    {
	while (!toParts[i].empty()) {
	    o_4b		to = toParts[i].remove() - lowId;
	    Link<Connection>	connection;

	    connection = new Persistent Connection(atom[i], atom[to]);
	    connection.unpinobj();
	}
	atom[i]->unpinobj();
    }

    delete atom;
    delete toParts;

    return count;
}
