/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/

#include <libc.h>
#include <string.h>
#include <stdio.h>

#include "oo7.h"
#include "BenchParams.h"
#include "GenParams.h"
#include "globals.h"

//////////////////////////////////////////////////////////////////
//
// Traverse  - DFS traverse module.  Upon reaching a base assembly,
// visit all referenced composite parts.  At each composite part,
// take an action that depends on which traversal variant has been
// requested.
//
///////////////////////////////////////////////////////////////////

extern o_4b	debugMode;
extern void	PrintOp(BenchmarkOp op);

//////////////////////////////////////////////////////////////////
//
// Module Method for Traversal 
//
//////////////////////////////////////////////////////////////////

o_4b Module::traverse(BenchmarkOp op)
{
    if (debugMode) {
        printf("Module::traverse(id = %d, op = ", id);
	PrintOp(op);
	printf(")\n");
    }

    // group read the base assemblies
    ::dom->greadobjs(assemblies, NULL);

    // now traverse the assembly hierarchy
    return designRoot->traverse(op);
}


//////////////////////////////////////////////////////////////////
//
// ComplexAssembly Method for Traversal
//
//////////////////////////////////////////////////////////////////

o_4b ComplexAssembly::traverse(BenchmarkOp op)
{
    if (debugMode) {
        printf("\tComplexAssembly::traverse(id = %d, op = ", id);
	PrintOp(op);
	printf(")\n");
    }

    // traverse each of the assembly's subassemblies
    o_4b	size = subAssemblies.size();
    o_4b	count = 0;

    for (o_word i = 0; i < size; i++ ) {
	count += subAssemblies[i]->traverse(op);
    }
    return count;
}


//////////////////////////////////////////////////////////////////
//
// BaseAssembly Method for Traversal 
//
//////////////////////////////////////////////////////////////////


o_4b BaseAssembly::traverse(BenchmarkOp op)
{
    o_4b	size = ComponentsPriv.size();
    o_4b	count = 0;

    if (debugMode) {
        printf("\t\tBaseAssembly::traverse(id = %d, op =  ", id);
	PrintOp(op);
	printf(" Privs = %d ", size);
	printf(")\n");
    }

    // establish iterator of private composite parts
    for (o_word i = 0; i < size; i++) {
	count += ComponentsPriv[i]->traverse(op);
    }
    return count;
}


//////////////////////////////////////////////////////////////////
//
// CompositePart Method for Traversal
//
//////////////////////////////////////////////////////////////////


o_4b CompositePart::traverse(BenchmarkOp op)
{
    if (debugMode) {
        printf("\t\t\tCompositePart::traverse(id = %d, op = ", id);
	PrintOp(op);
	printf(")\n");
    }

    // do DFS of the composite part's atomic part graph
    if ((op >= Trav1) && (op <= Trav3c)) {
	// group read the parts
	o_lockmode		lockMode;

	switch(op)
	{
	case Trav1:
	case Trav1WW:
	case Trav2a:
	case Trav3a:
	    lockMode = O_CONVERT_LOCK(RLOCK);
	    break;
	case Trav2b:
	case Trav2c:
	case Trav3b:
	case Trav3c:
	    lockMode = O_CONVERT_LOCK(WLOCK);
	    break;
	}
	::dom->greadobjs(parts, NULL, lockMode);

        // do parameterized DFS of atomic part graph
        PartIdSet	visitedIds;

        return rootPart->traverse(op, visitedIds);
    } else if (op == Trav4) {
        // search document text for a certain character
        return document->searchText('I');
    } else if (op == Trav5do) {
        // conditionally change initial part of document text
        return document->replaceText("I am", "This is");
    } else if (op == Trav5undo) {
        // conditionally change back initial part of document text
        return document->replaceText("This is", "I am");
    } else if (op == Trav6) {
	// visit the root part only (it knows how to handle this)
        PartIdSet	visitedIds;

        return rootPart->traverse(op, visitedIds);
    } else {
	// composite parts don't respond to other traversals
	printf("*** CompositePart::PANIC -- illegal traversal!!! ***\n");
	return 0;
    }
}


//////////////////////////////////////////////////////////////////
//
// AtomicPart Method for Traversal
//
//////////////////////////////////////////////////////////////////


o_4b AtomicPart::traverse(BenchmarkOp op, PartIdSet& visitedIds)
{
    if (debugMode) {
        printf("\t\t\tAtomicPart::traverse(id = %d, op = ", id);
	PrintOp(op);
	printf(")\n");
    }

    o_word	i;
    o_4b	count=0; // was 1 in Version 1. Why???

    switch (op) {
	case Trav1:
	    // just examine the part
            count += 1;
            DoNothing();
	    break;

        case Trav2a:
	    // swap X and Y if first part
	    if (visitedIds.empty()) { 
                swapXY(); 
                count += 1;
            }
	    break;

        case Trav2b:
	    // swap X and Y
            swapXY();
            count += 1;
	    break;

        case Trav2c:
	    // swap X and Y repeatedly
            for (i = 0; i < UpdateRepeatCnt; i++) { 
               swapXY(); 
               count += 1;
            }
	    break;

        case Trav3a:
	    // toggle date if first part
	    if (visitedIds.empty()) { 
                toggleDate(); 
                count += 1;
            }
	    break;

        case Trav3b:
	    // toggle date
            toggleDate();
            count += 1;
	    break;

        case Trav3c:
	    // toggle date repeatedly
            for (i = 0; i < UpdateRepeatCnt; i++) { 
               toggleDate();
               count += 1;
            }
	    break;

	case Trav6:
	    // examine only the root part
            count += 1;
            DoNothing();
	    return count;

        default:
	    // atomic parts don't respond to other traversals
	    printf("*** AtomicPart::PANIC -- illegal traversal!!! ***\n");

    }
    // now, record the fact that we've visited this part
    visitedIds.insert(id);

    // finally, continue with a DFS of the atomic parts graph
    // establish iterator over connected parts
    o_4b	size = to.size();

    for (i = 0; i < size; i++) {
	AtomicPart*	atomicPart = to[i]->to;

	if (!visitedIds.contains(atomicPart->id)) {
	    count += atomicPart->traverse(op, visitedIds);
	}
    }
    unpinobj();
    return count;
}
